<?php

namespace App\Services;

use App\Models\Station;
use App\Models\WeatherData;
use Carbon\CarbonImmutable;
use Carbon\CarbonInterface;
use DateTimeInterface;
use Illuminate\Support\Arr;
use PhpOffice\PhpSpreadsheet\Writer\Xls as XlsWriter;

class ExportData
{
    public $excelPath = null;

    public function __construct()
    {
        $this->excelPath = storage_path('app/xls');
    }

    public function process(CarbonInterface $startDate = null)
    {
        $startDate = $startDate ?: CarbonImmutable::now()->subDays(1);
        $endDate = $startDate->clone()->addDay();

        $data = $this->getData($startDate->startOfDay(), $endDate->endOfDay());

        // $this->makeDailyExport($data);

        $this->makeMonthyExport($data);
    }

    public function getData(DateTimeInterface $from, DateTimeInterface $to)
    {
        $weather_data = WeatherData::whereBetween('software_channel_datetime', [$from, $to])
            ->orderBy('software_channel_sn', 'desc')
            ->orderBy('software_channel_datetime', 'asc')
            ->cursor();

        $exportData = [];

        //make the export array
        foreach ($weather_data as $d) {
            $exportData[$d->software_channel_sn][$d->software_channel_datetime][$d->software_channel_name] = $d->software_channel_value;
        }

        return $exportData;
    }

    public function makeMonthyExport($exportData)
    {
        $stations = Station::all()->keyBy('sn')->toArray();

        foreach ($exportData as $currentStation => $snData) {

            try {
                $code = $stations[$currentStation]['code'];
            } catch (\Throwable $e) {
                if ($currentStation === '041A0272') $code = 'CRF-4';
                else {
                    echo "Error '$currentStation' Not Found: {$e->getMessage()}";
                    continue;
                }
            }
            $code_sufix = match (substr($code, 0, 3)) {
                'CRF' => '1-CORFU',
                'KEF' => '2-CEPHALONIA',
                'ZKT' => '3-ZAKYNTHOS',
                default => null
            };
            $date = new CarbonImmutable(array_key_first($snData));
            $date_format = strtoupper($date->format('n-M-Y'));

            $excelPath = "{$this->excelPath}/Monthly/{$date->format("Y-m")}";
            if (!file_exists($excelPath)) mkdir($excelPath, 0755, true);

            $filenames = [
                "A0_IW_per MONTH_STATISTICS_{$date->format('n-M')}.xls",
                "A1_IW_per DAY data for_{$code}_{$date_format}.xls",
                "A2_IW_Monthly_Statistics_for_{$date_format}.xls",
                "A3_IW_Monthly_ANOMALIES_{$date_format}.xls",
                "A4_IW_TEMP STATS_for_{$code_sufix}.xls",
                "A5_IW_RAIN STATS_for_{$code_sufix}.xls",
            ];

            $spreadsheets = Arr::map($filenames, function ($file, $i) use ($excelPath) {
                $xlsReader = \PhpOffice\PhpSpreadsheet\IOFactory::createReader("Xls");

                $paths = [
                    "$excelPath/$file",
                    "$this->excelPath/Templates/$file",
                    "$this->excelPath/Templates/A$i.xls"
                ];

                foreach ($paths as $path) {
                    if (file_exists($path)) return $xlsReader->load($path);
                }

                return null;
            });
            $sheets = Arr::map($spreadsheets, fn ($spreadsheet) => $spreadsheet?->getActiveSheet());

            // Make A1 Export (Daily)
            $data = collect($snData);

            if ($data->count('WD') > 0) {
                $WDdata = $data->map(
                    fn ($val) => isset($val['WD']) ? [
                        'WD_x' => -1 * round(sin(pi() * $val['WD'] / 180), 5),
                        'WD_y' => -1 * round(cos(pi() * $val['WD'] / 180), 5),
                    ] : []
                );

                $Vx = $WDdata->avg('WD_x');
                $Vy = $WDdata->avg('WD_y');

                $WDavg = match (true) {
                    $Vx > 0 => atan($Vy / $Vx) + 180,
                    $Vx < 0 => $Vy < 0 ? atan($Vy / $Vx) : (atan($Vy / $Vx) + 360),
                    default => $Vy < 0 ? 0 : 180,
                };
            }

            $i = intval($date->format('j')) + 1;
            $sheets[1]
                ->setCellValue("B$i", WeatherData::getDN($date->startOfDay()))     // DN	
                ->setCellValue("C$i", $data->sum('RT'))     // Rain	
                ->setCellValue("D$i", $data->max('RT'))     // RRmax	
                ->setCellValue("E$i", $data->avg('WS'))     // VELav	
                ->setCellValue("F$i", $data->max('MaxWS'))  // GUST	
                ->setCellValue("G$i", $WDavg ?? null)       // DIR	 
                ->setCellValue("H$i", $data->avg('AP'))     // PRES	
                ->setCellValue("I$i", $data->avg('AT'))     // Tav	
                ->setCellValue("J$i", $data->min('AT'))     // Tmin	
                ->setCellValue("K$i", $data->max('AT'))     // Tmax	
                ->setCellValue("L$i", ($data->max('AT') - $data->min('AT')))   // Trange	
                ->setCellValue("M$i", $data->avg('RH'))     // RH	
                ->setCellValue("N$i", $data->max('RH'))     // RHmax	
                ->setCellValue("O$i", $data->max('SR'));    // MaxIrradiance

            // Make A2 Export (Monthly)
            $i = 0;
            foreach ($sheets[2]->getRowIterator() as $row) {
                $r = $row->getRowIndex();
                $cell = $sheets[2]->getCell("A$r");

                if ($cell->getValue() === $code) {
                    $i = $cell->getRow();
                    break;
                }
            }

            if (empty($i)) continue;

            $title = "{$date->locale('el')->monthName} {$date->format('Y')} - ΚΛΙΜΑΤΙΚΟΙ ΔΕΙΚΤΕΣ   ΑΝΑ ΣΤΑΘΜΟ";
            $sheets[2]->setCellValue("B1", $title);
            $sheets[2]->setCellValue("B$i", \PhpOffice\PhpSpreadsheet\Shared\Date::PHPToExcel($date->startOfMonth()));
            $i = $i + 2;
            foreach (range(35, 40) as $row) {
                foreach (range('C', 'N') as $col) {
                    $c = chr(ord($col) - 1);
                    $sheets[2]->setCellValue("$c$i", $this->getValue($sheets[1], "$col$row"));
                }
                $i++;
            }

            // Get A0 Values for Anomalies
            $i = 0;
            foreach ($sheets[0]->getRowIterator() as $row) {
                $r = $row->getRowIndex();
                $cell = $sheets[2]->getCell("A$r");

                if ($cell->getValue() === $code) {
                    $i = $cell->getRow();
                    break;
                }
            }

            $meanRainPrecipitation = $this->getValue($sheets[0], "D" . ($i + 7));
            $minTemprature = $this->getValue($sheets[0], "I" . ($i + 3));
            $avgTemprature = $this->getValue($sheets[0], "I" . ($i + 4));
            $maxTemprature = $this->getValue($sheets[0], "I" . ($i + 5));
            $maxGust = $this->getValue($sheets[0], "F" . ($i + 5));

            // Get A4 Values for Anomalies
            $i = 0;
            $TEMP1900 = [];
            if (!empty($sheets[4])) {
                foreach ($sheets[4]->getColumnIterator() as $col) {
                    $cell = $sheets[4]->getCell("{$col->getColumnIndex()}2");
                    if (str_starts_with($cell->getValue(), strtoupper($date->format('M')))) {
                        $i = $cell->getColumn();
                        break;
                    }
                }

                foreach (range(13, 19) as $c) {
                    $TEMP1900[$c] = $this->getValue($sheets[4], "{$i}{$c}");
                }
            }

            // Get A5 Values for Anomalies
            $i = 0;
            $RHP1955 = null;
            $RHP1900 = null;
            if (!empty($sheets[5])) {
                foreach ($sheets[5]->getColumnIterator() as $col) {
                    $cell = $sheets[5]->getCell("{$col->getColumnIndex()}1");
                    if (str_starts_with($cell->getValue(), strtoupper($date->format('M')))) {
                        $i = $cell->getColumn();
                        break;
                    }
                }

                $RHP1955 = $this->getValue($sheets[5], "{$i}5");
                $RHP1900 = $this->getValue($sheets[5], "{$i}16");
            }

            // Make A3 Export (Anomalies)
            $i = null;
            foreach ($sheets[3]->getColumnIterator() as $col) {
                $cell = $sheets[3]->getCell("{$col->getColumnIndex()}1");

                if ($cell->getValue() === $code) {
                    $i = $cell->getColumn();
                    break;
                }
            }
            if (empty($i)) continue;

            $MRain = $this->getValue($sheets[1], "C36");
            $rainMax = $this->getValue($sheets[1], "C38");
            $RRMax   = $this->getValue($sheets[1], "D38");
            $tempMin = $this->getValue($sheets[1], "J37");
            $tempAvg = $this->getValue($sheets[1], "I39");
            $tempMax = $this->getValue($sheets[1], "K38");
            $gustMax = $this->getValue($sheets[1], "F38");

            $rainDays = 0;
            foreach (range(2, 32) as $row) {
                if ($this->getValue($sheets[1], "C$row")) $rainDays++;
            }

            $sheets[3]
                ->setCellValue("{$i}2", $MRain)
                ->setCellValue("{$i}3", $meanRainPrecipitation ? (100 * ($MRain / $meanRainPrecipitation)) : null) # LRHP
                ->setCellValue("{$i}4", $tempAvg)
                ->setCellValue("{$i}5", ($tempAvg - $avgTemprature))
                ->setCellValue("{$i}6", $tempMin)
                ->setCellValue("{$i}7", ($tempMin - $minTemprature))
                ->setCellValue("{$i}8", $tempMax)
                ->setCellValue("{$i}9", ($tempMax - $maxTemprature))
                ->setCellValue("{$i}10", $gustMax)
                ->setCellValue("{$i}11", ($gustMax - $maxGust))
                ->setCellValue("{$i}12", $RRMax)
                ->setCellValue("{$i}13", $rainMax)
                ->setCellValue("{$i}14", $rainDays)
                ->setCellValue("{$i}15", null) # empty line
                ->setCellValue("{$i}16", $RHP1955 ? (100 * ($MRain - $RHP1955) / $RHP1955) : null)
                ->setCellValue("{$i}17", $RHP1900 ? (100 * ($MRain - $RHP1900) / $RHP1900) : null);

            if (!empty($TEMP1900)) {
                $sheets[3]
                    ->setCellValue("{$i}19", ($tempMax - $TEMP1900[13]))
                    ->setCellValue("{$i}20", ($tempMax - $TEMP1900[14]))
                    ->setCellValue("{$i}21", ($tempAvg - $TEMP1900[15]))
                    ->setCellValue("{$i}22", ($tempAvg - $TEMP1900[16]))
                    ->setCellValue("{$i}23", ($tempAvg - $TEMP1900[17]))
                    ->setCellValue("{$i}24", ($tempMin - $TEMP1900[18]))
                    ->setCellValue("{$i}25", ($tempMin - $TEMP1900[19]));
            }

            // Save exports and close sheets
            foreach ($spreadsheets as $key => $spreadsheet) {
                // Files A1, A2, A3
                if ($key >= 1 && $key <= 3) (new XlsWriter($spreadsheet))->save("$excelPath/$filenames[$key]");
                $spreadsheet?->disconnectWorksheets();
                $spreadsheet?->garbageCollect();
            }

            unset($spreadsheets, $sheets);
        }
    }

    public function makeDailyExport($exportData)
    {
        foreach ($exportData as $currentStation => $snData) {

            $spreadsheet = new \PhpOffice\PhpSpreadsheet\Spreadsheet();

            $spreadsheet->getProperties()->setCreator("Ionian Weather")
                ->setLastModifiedBy("Ionian Weather")
                ->setTitle("Ionian weather")
                ->setSubject("Ionian weather")
                ->setDescription("Ionian weather")
                ->setKeywords("Ionian weather")
                ->setCategory("Ionian weather");

            $objPHPExcelSheet = $spreadsheet->getActiveSheet();

            $i = 1;
            $objPHPExcelSheet->setTitle($currentStation)
                ->setCellValue("A$i", 'Date')
                ->setCellValue("B$i", 'Time')
                ->setCellValue("C$i", 'DN')
                ->setCellValue("D$i", 'Rain')
                ->setCellValue("E$i", 'Vel avg')
                ->setCellValue("F$i", 'Vel min')
                ->setCellValue("G$i", 'Vel max')
                ->setCellValue("H$i", 'Vel sdv')
                ->setCellValue("I$i", 'Dir avg')
                ->setCellValue("J$i", 'Dir min')
                ->setCellValue("K$i", 'Dir max')
                ->setCellValue("L$i", 'Dir sdv')
                ->setCellValue("M$i", 'Pressure')
                ->setCellValue("N$i", 'Temperature')
                ->setCellValue("O$i", 'RH')
                ->setCellValue("P$i", 'Solar Irradiance')
                ->setCellValue("Q$i", 'UV-A')
                ->setCellValue("R$i", 'UV-B')
                ->setCellValue("S$i", 'PM1')
                ->setCellValue("T$i", 'PM2.5')
                ->setCellValue("U$i", 'PM4')
                ->setCellValue("V$i", 'PM10')
                ->setCellValue("W$i", 'Total')
                ->setCellValue("X$i", 'MS')
                ->setCellValue("Y$i", 'FP')
                ->setCellValue("Z$i", 'ES')
                ->setCellValue("AA$i", 'MVAP')
                ->setCellValue("AB$i", 'PVAP')
                ->setCellValue("AC$i", 'SH')
                ->setCellValue("AD$i", 'AH')
                ->setCellValue("AE$i", 'K')
                ->setCellValue("AF$i", 'TDEW')
                ->setCellValue("AG$i", 'TVIR')
                ->setCellValue("AH$i", 'RHO')
                ->setCellValue("AI$i", 'APWR')
                ->setCellValue("AJ$i", 'EAN')
                ->setCellValue("AK$i", 'SEN');

            foreach (range('A', 'Z') as $columnID) {
                $objPHPExcelSheet->getColumnDimension($columnID)->setAutoSize(true);
                $objPHPExcelSheet->getColumnDimension('A' . $columnID)->setAutoSize(true);
            }

            $currentDate = CarbonImmutable::now();
            foreach ($snData as $stationDate => $dateData) {
                $currentDate = new CarbonImmutable($stationDate);

                $i++;
                $objPHPExcelSheet
                    ->setCellValue("A$i",  $currentDate->format('Y-m-d'))
                    ->setCellValue("B$i", $currentDate->format('H:i'))
                    ->setCellValue("C$i",  WeatherData::getDN($currentDate));

                $eData = [];

                foreach ($dateData as $channel => $value) {
                    $stVal = round($value, 4);
                    switch ($channel) {
                        case 'RT':
                            $eData['rain'] = $stVal;
                            break;
                        case 'AT':
                            $eData['temprature'] = $stVal;
                            break;
                        case 'WS':
                            $eData['windSpeed'] = $stVal;
                            break;
                        case 'MaxWS':
                            $eData['MaxWS'] = $stVal;
                            break;
                        case 'MinWS':
                            $eData['MinWS'] = $stVal;
                            break;
                        case 'SDWS':
                            $eData['SDWS'] = $stVal;
                            break;
                        case 'WD':
                            $eData['windDir'] = $stVal;
                            break;
                        case 'MinWD':
                            $eData['MinWD'] = $stVal;
                            break;
                        case 'MaxWD':
                            $eData['MaxWD'] = $stVal;
                            break;
                        case 'SDWD':
                            $eData['MaxWD'] = $stVal;
                            break;
                        case 'AP':
                            $eData['pressure'] = $stVal;
                            break;
                        case 'RH':
                            $eData['humidity'] = $stVal;
                            break;
                        case 'SR':
                            $eData['irradiance'] = $stVal;
                            break;
                        case 'PM1':
                            $eData['PM1'] = $stVal;
                            break;
                        case 'PM2.5':
                            $eData['PM2.5'] = $stVal;
                            break;
                        case 'PM4':
                            $eData['PM4'] = $stVal;
                            break;
                        case 'PM10':
                            $eData['PM10'] = $stVal;
                            break;
                        case 'TP':
                            $eData['total'] = $stVal;
                            break;
                        case 'UVA':
                            $eData['uva'] = $stVal;
                            break;
                        case 'UVB':
                            $eData['uvb'] = $stVal;
                            break;

                        default:
                            $eData['default'] = $stVal;
                    }
                }

                if (!empty($eData)) {
                    //make the secondary parameters
                    $T = $eData['temprature'] ?? 0;
                    $P = $eData['pressure'] ?? 0;
                    $RH = $eData['humidity'] ?? 0;
                    $u = $eData['windSpeed'] ?? 0;

                    $eData['MS'] = 3.5346 + 0.2457 * $T + 0.0091 * pow($T, 2) + 0.0002 * pow($T, 3);
                    $MS = $eData['MS'];

                    $eData['MVAP'] = $RH * $MS / 100;

                    if ($P != 0) {
                        $eData['FP'] = 1.0016 + 3.15 * pow(10, -6) * $P - 0.074 / $P;
                        $FP = $eData['FP'];

                        $a = 7.5;
                        $b = 237.3;
                        $exp = ($a * $T) / ($b + $T);
                        $eData['ES'] = 6.112 * $FP * pow(10, $exp);
                        $ES = $eData['ES'];

                        $eData['PVAP'] = $RH * $ES / 100;
                        $PVAP = $eData['PVAP'];

                        $eData['SH'] = 0.622 * $PVAP / ($P - $PVAP);
                        $SH = $eData['SH'];

                        $eData['AH'] = (216.688 * $PVAP) / ($T + 273);

                        $K = log($PVAP / (6.112 * $FP));
                        $eData['K'] = $K;
                        $eData['TDEW'] = 243.12 * $K / (17.62 - $K);

                        $eData['TVIR'] = (1 + 0.61 * $SH) * $T;
                        $TVIR = $eData['TVIR'];

                        $Rdry = 287;
                        $eData['RHO'] = ($P * 100) / ($Rdry * ($TVIR + 273));
                        $RHO = $eData['RHO'];

                        $eData['APWR'] = $RHO * pow($u, 3) / 2;
                    }

                    //save the line
                    $objPHPExcelSheet
                        ->setCellValue("D$i", $eData['rain'] ?? '')
                        ->setCellValue("E$i", $eData['windSpeed'] ?? '')
                        ->setCellValue("F$i", $eData['MinWS'] ?? '')
                        ->setCellValue("G$i", $eData['MaxWS'] ?? '')
                        ->setCellValue("H$i", $eData['SDWS'] ?? '')
                        ->setCellValue("I$i", $eData['windDir'] ?? '')
                        ->setCellValue("J$i", $eData['MinWD'] ?? '')
                        ->setCellValue("K$i", $eData['MaxWD'] ?? '')
                        ->setCellValue("L$i", $eData['SDWD'] ?? '')
                        ->setCellValue("M$i", $eData['pressure'] ?? '')
                        ->setCellValue("N$i", $eData['temprature'] ?? '')
                        ->setCellValue("O$i", $eData['humidity'] ?? '')
                        ->setCellValue("P$i", $eData['irradiance'] ?? '')
                        ->setCellValue("Q$i", $eData['uva'] ?? '')
                        ->setCellValue("R$i", $eData['uvb'] ?? '')
                        ->setCellValue("S$i", $eData['PM1'] ?? '')
                        ->setCellValue("T$i", $eData['PM2.5'] ?? '')
                        ->setCellValue("U$i", $eData['PM4'] ?? '')
                        ->setCellValue("V$i", $eData['PM10'] ?? '')
                        ->setCellValue("W$i", $eData['total'] ?? '')
                        ->setCellValue("X$i", $eData['MS'] ?? '')
                        ->setCellValue("Y$i", $eData['FP'] ?? '')
                        ->setCellValue("Z$i", $eData['ES'] ?? '')
                        ->setCellValue("AA$i", $eData['MVAP'] ?? '')
                        ->setCellValue("AB$i", $eData['PVAP'] ?? '')
                        ->setCellValue("AC$i", $eData['SH'] ?? '')
                        ->setCellValue("AD$i", $eData['AH'] ?? '')
                        ->setCellValue("AE$i", $eData['K'] ?? '')
                        ->setCellValue("AF$i", $eData['TDEW'] ?? '')
                        ->setCellValue("AG$i", $eData['TVIR'] ?? '')
                        ->setCellValue("AH$i", $eData['RHO'] ?? '')
                        ->setCellValue("AI$i", $eData['APWR'] ?? '');
                }
            }

            $excelPath = "{$this->excelPath}/Daily/{$currentDate->format("Y-m-d")}";
            if (!file_exists($excelPath)) mkdir($excelPath, 0755, true);

            (new XlsWriter($spreadsheet))->save("$excelPath/$currentStation.xls");

            $spreadsheet->disconnectWorksheets();
            $spreadsheet->garbageCollect();
            unset($spreadsheet, $objPHPExcelSheet);
        }
    }

    private function getValue($sheet, $cell)
    {
        return $sheet->getCell($cell)->getCalculatedValue();
    }
}
